use game::*;
use bitboard::*;
use movegen::*;

use std::i32;

pub type PosValue = i32;

pub const MIN_VALUE: PosValue = i32::MIN + 1;
pub const MAX_VALUE: PosValue = i32::MAX;

const MATE_VALUE: PosValue = 1i32 << 26;
const MATE_CUTOFF: PosValue = 1i32 << 24;

pub fn mate() -> PosValue {
    mate_in_p1(1)
}

/**
 * constructs a PosValue that indicates that from a given position mate
 * can be achieved in turns-1 moves (not halfmoves)
 */
pub fn mate_in_p1(turns: i32) -> PosValue {
    if turns < 0 {
        return -mate_in_p1(-turns);
    }
    else {
        return MATE_VALUE - turns;
    }
}

/**
 * returns Some(turns) if the pos_val indicates that mate
 * can be reached in turns-1 moves
 * 
 * turns is negative if the moving side is getting mated
 * in turns moves
 */
pub fn is_mate_in_p1(pos_val: PosValue) -> Option<i32> {
    if pos_val > MATE_CUTOFF && pos_val <= MATE_VALUE {
        Some(-pos_val + MATE_VALUE)
    }
    else if pos_val < -MATE_CUTOFF {
        Some(-pos_val - MATE_VALUE)
    }
    else {
        None
    }
}

pub fn increase_mate_in(mut val: PosValue) -> PosValue {
    if let Some(turns) = is_mate_in_p1(val) {
        if turns < 0 {
            val = mate_in_p1(turns - 1);
        }
        if turns > 0 {
            val = mate_in_p1(turns + 1);
        }
        if turns == 0 {
            if val < 0 {
                val = mate_in_p1(-1);
            }
            else {
                val = mate_in_p1(1);
            }
        }
    }
    val
}

pub fn decrease_mate_in(mut val: PosValue) -> PosValue {
    if let Some(turns) = is_mate_in_p1(val) {
        if turns < 0 {
            val = mate_in_p1(turns + 1);
        }
        if turns > 0 {
            val = mate_in_p1(turns - 1);
        }
    }
    val
}


fn side_value(game: &Game, side: Side) -> u32 {
    let adv_pawn_mask = match side { WHITE => ROW_7, BLACK => ROW_2 };
    let semi_adv_pawn_mask = match side { WHITE => ROW_6, BLACK => ROW_3 };
    let advanced_pawns = (game.get_piece(PAWN, side) & adv_pawn_mask).count_ones() * 200;
    let semi_advanced_pawns = (game.get_piece(PAWN, side) & semi_adv_pawn_mask).count_ones() * 50;
    advanced_pawns + semi_advanced_pawns
}

fn knight_value(game: &Game, side: Side) -> PosValue {
    let knights = game.get_piece(KNIGHT, side);
    let ks = BitboardIterator(knights);
    let mut k_attacks: u32 = 0;

    for k in ks {
        let targets = get_knight_targets(square(k));
        k_attacks += targets.count_ones() * 8;
    }

    let num_opp_pawns = game.get_piece(PAWN, !side).count_ones() as PosValue;
    let num_knights = game.get_piece(KNIGHT, side).count_ones() as PosValue;

    let center = (ROW_4 | ROW_5) & (FILE_D | FILE_E);
    let center_knights = (game.get_piece(KNIGHT, side) & center).count_ones() as PosValue;
    
    center_knights * 20 + k_attacks as PosValue + ((num_knights * 75 * num_opp_pawns) / 8) as PosValue
}

fn material_value(game: &Game, side: Side) -> PosValue {
    (game.get_piece(PAWN, side).count_ones() * 100
        + game.get_piece(KNIGHT, side).count_ones() * 220
        + game.get_piece(BISHOP, side).count_ones() * 320
        + game.get_piece(ROOK, side).count_ones() * 400
        + game.get_piece(QUEEN, side).count_ones() * 700) as PosValue
}

fn king_safety(game: &Game, side: Side) -> PosValue {
    let king = game.get_piece(KING, side);
    let area = north_one(king)
             | south_one(king)
             | east_one(king)
             | west_one(king)
             | northeast_one(king)
             | northwest_one(king)
             | southwest_one(king)
             | southeast_one(king);
    
    let guards = game.get_all_side(side) & area;
    let attackers = game.get_all_side(!side) & area;

    let kneighbors = guards.count_ones() as PosValue * 5 - attackers.count_ones() as PosValue * 15;

    let enemies = game.get_all_side(!side);
    let kingrow = indices_from_square(square(king)).1;

    let row_opponents = enemies & (ROW_1 << (kingrow * 8));
    let dang_row_opponents = row_opponents & (ROW_1 | ROW_2 | ROW_7 | ROW_8);
    let danger_value = (dang_row_opponents.count_ones() * 50) as PosValue;

    return kneighbors + danger_value;
}

fn king_there(game: &Game, side: Side) -> PosValue {
    if game.get_piece(KING, side) != 0 { 10000 } else { 0 }
}

pub fn evaluate(game: &Game) -> PosValue {
    let sv = side_value(game, game.turn) as PosValue - side_value(game, !game.turn) as PosValue;
    let material_value_us = material_value(game, game.turn);
    let material_value_them = material_value(game, !game.turn);

    let mat_val = (material_value_us) - (material_value_them) as PosValue;

    let kv = knight_value(game, game.turn) - knight_value(game, !game.turn);
    let king_safety = king_safety(game, game.turn) - king_safety(game, !game.turn);
    let king_there = king_there(game, game.turn) - king_there(game, !game.turn);

    let mge = sv + kv + king_safety + king_there;
    let lge = late_game_eval(game);

    let lateness = game_lateness(game);

    let ge = (mge * (100 - lateness) + lge * lateness) / 100;


    return mat_val + ge;
}

pub fn game_lateness(game: &Game) -> i32 {
    let all_pieces = game.get_all_side(WHITE) | game.get_all_side(BLACK);
    let n_pieces = all_pieces.count_ones();
    let captured_pieces = 32 - n_pieces;

    if captured_pieces > 15 {
        let lateness = (captured_pieces - 15) * 100 / 10;
        return std::cmp::min(lateness as _, 100);
    }
    return 0;
}

pub fn late_game_eval(game: &Game) -> PosValue {
    let pp = pawn_push_value(game, game.turn) - pawn_push_value(game, !game.turn);
    let bp = blocked_pawns(game, game.turn) - blocked_pawns(game, !game.turn);

    pp + bp
}

pub fn pawn_push_value(game: &Game, side: Side) -> PosValue {
    let val = if side == WHITE {
        (game.get_piece(PAWN, side) & ROW_3).count_ones() * 10
        + (game.get_piece(PAWN, side) & ROW_4).count_ones() * 20
        + (game.get_piece(PAWN, side) & ROW_5).count_ones() * 30
        + (game.get_piece(PAWN, side) & ROW_6).count_ones() * 40
        + (game.get_piece(PAWN, side) & ROW_7).count_ones() * 50
    }
    else {
        (game.get_piece(PAWN, side) & ROW_6).count_ones() * 10
        + (game.get_piece(PAWN, side) & ROW_5).count_ones() * 20
        + (game.get_piece(PAWN, side) & ROW_4).count_ones() * 30
        + (game.get_piece(PAWN, side) & ROW_3).count_ones() * 40
        + (game.get_piece(PAWN, side) & ROW_2).count_ones() * 50
    };
    val as PosValue
}

pub fn blocked_pawns(game: &Game, side: Side) -> PosValue {
    let pawns = game.get_piece(PAWN, side);
    let pushed_pawns = if side == WHITE {
        north_one(pawns)
    } else {
        south_one(pawns)
    };

    let blocked_pawns = pushed_pawns & game.get_all_side(!side);

    blocked_pawns.count_ones() as _
}



#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_mate_values() {
        let mate_in_three = mate_in_p1(3);
        let mate_in_four = mate_in_p1(4);

        assert!(mate_in_three > mate_in_four);
        assert!(mate_in_three > mate_in_four);
    }
}

